/**
 * @author prat@cines.fr
 * last maj : 2010/08/02
 * */

package fr.cines.pac.format;

import java.io.File;
import java.io.IOException;
import java.util.*;

import org.apache.log4j.Logger;
import org.apache.log4j.xml.DOMConfigurator;
import org.jdom.input.SAXBuilder;
import org.jdom.*;

import fr.cines.pac.formatValidator.exception.FormatValidatorException;
import fr.cines.pac.formatValidator.tools.DroidToolSingleton;
import fr.cines.pac.formatValidator.util.XsdSchemaSaxValidator;

public class Config {
	
	public final static String codeVersion ="2.3";
	/** valeur de designant la version du code java de l'application */
		
	Element eRoot;
	/** valeur de la balise xml des references hypertextes */
	public final static String A_HREF="href";
	
	/** valeur de la balise xml de reference des chemins du système de fichier */
	public final static String A_PATH="path";

	/** valeur de la balise xml designant le nom de l'application */
	public final static String E_NAME				="name";
	
	/** valeur de la balise xml designant la version de l'application */
	public final static String E_VERSION			="version";
		
	/** valeur de la balise xml designant le répertoire temporaire */
	public final static String E_TMP				="tmp";
	
	/** valeur de la balise xml designant le fichier de configuration des logs */
	public final static String E_LOG4J_CONF			="log4jConf";
	
	/** valeur de la balise xml designant le fichier de configuration de l'outil DROID */
	public final static String E_DROID_CONF			="droidConf";
	
	/** valeur de la balise xml designant le fichier de configuration de l'outil JHOVE */
	public final static String E_JHOVE_CONF			="jhoveConf";
	
	/** valeur de la balise xml designant l'outil mplayer */
	public final static String E_MPLAYER			="mplayer";
	
	/** valeur de la balise xml designant l'outil mediaInfo */
	public final static String E_MEDIAINFO			="mediainfo";
	
	/** valeur de la balise xml designant l'outil imageMagick */
	public final static String E_IMAGICK_IDENTIFY	="imagickIdentify";
	
	/** valeur de la balise xml designant l'outil imagickConvert de imageMagick */
	public final static String E_IMAGICK_CONVERT	="imagickConvert";
	
	/** valeur de la balise xml designant le message par defaut du validateur ( si pas de message d'erreur) */
	public final static String E_DEF_VALMESS		="defaultValidatorMessage";
	
	/** valeur de la balise xml designant le message par defaut du validateur ( si pas de message d'erreur) */
	public final static String E_MAX_SIZE_MESSAGE		="maxSizeMessage";
	
	/** valeur de la balise xml designant la balise relative aux formats */
	public final static String E_FORMAT				="format";
	
	/** valeur de la balise xml designant la balise relative aux outils */
	public final static String E_TOOL				="tool";

	/** instance unique de l'object de configuration représenté par cette classe*/
	private static Config instance;

	/** logger utilisé pour la configuration*/
	Logger logger = null;

	private File	
	home,					// repertoire racine de l'application
	xmlFile,				// fichier xml de configuration		
	xsdFile,				// fichier xsd de configuration
	tmpDir,					// repertoir temporaire de l'application
	log4jConf,				// fichier de configuration des logs
	droidConf,				// fichier de configuration de l'outil DROID
	jhoveConf, 				// fichier de configuration de l'outil JHOVE
	mPlayerPath,			// chemin de l'outil mplayer
	mediaInfoPath,			// chemin de l'outil mediainfo
	iMagickIdentifyPath,	// chemin de l'outil imageMagick
	iMagickConvertPath;		// chemin de l'outil ImageCOnvert d'imageMagick

	private String	
	version,				// version de l'application
	name,					// nom de l'application
	defaultValidatorMessage,// message par defaut du validateur 
	homeJavaProperty;		// propriété java designant le repertoire racine de l'application		

	private int maxSizeMessage;
	
	ArrayList<FormatConf> 
	formatConfigList, 		// liste des objets de configuration de format
	formatConfigTestList, 	// liste des objets de configuration de format en test
	formatConfigProdList; 	// liste des objets de configuration de format en production

	static final String JAXP_SCHEMA_LANGUAGE = "http://java.sun.com/xml/jaxp/properties/schemaLanguage";

	static final String W3C_XML_SCHEMA = "http://www.w3.org/2001/XMLSchema";

	static final String JAXP_SCHEMA_SOURCE = "http://java.sun.com/xml/jaxp/properties/schemaSource"; 

	/**
	 *	0 si pas de controle effectué
	 *	1 si configuration est ok.
	 *	-1 si la configuration ests ko;
	 **/
	int check =0;

	/**
	 *controle de la configuration de l'application
	 **/
	public void check() throws FormatValidatorException{
		try{
			if(this.getHomeJavaProperty()==null) throw new Exception("java parameter [-Dfvhome=repertoire racine] is not set");
			if(!this.getHome().isDirectory()) throw new Exception("java parameter [-Dfvhome=repertoire racine] is not a directory");
			canRead(this.getXmlFile());
			canRead(this.getXsdFile());
			canRead(this.getTmpDir());
			canRead(this.getLog4jConf());
			canRead(this.getDroidConf());
			canRead(this.getJhoveConf());		
			canRead(this.getMPlayerPath());		
			canRead(this.getMediaInfoPath());		
			canRead(this.getIMagickIdentifyPath());	
			canRead(this.getIMagickConvertPath());	
			XsdSchemaSaxValidator.validateXml(xsdFile.getAbsolutePath(), xmlFile.getAbsolutePath());

			//controle de droid par initialisation
			DroidToolSingleton.initDroidSingleton();
		}catch(Exception e){
			this.check=-1;
			throw new FormatValidatorException(e);
		}
		this.check=1;
	}

	/**
	 *  test si l'application à déjà effectué un controle positif ou négatif.
	 **/
	public boolean isCheck(){
		return this.check!=0;
	}

	/**
	 * construction de l'oject config avec initialisations des varaibles necessaire
	 * au fonctionnement de l'application.
	 * */
	public Config(){

		System.setProperty("javax.xml.validation.SchemaFactory:http://relaxng.org/ns/structure/1.0", "org.iso_relax.verifier.jaxp.validation.RELAXNGSchemaFactoryImpl");
		System.setProperty("org.iso_relax.verifier.VerifierFactoryLoader","com.sun.msv.verifier.jarv.FactoryLoaderIm");

		System.setProperty("java.io.tmpdir",this.getTmpDir().getAbsolutePath());
		
		// initLog4j 
		DOMConfigurator.configure(this.getLog4jConf().getAbsolutePath());
		logger = Logger.getLogger(this.getClass().getName());
		logger.info("init "+this.getName()+" v"+Config.codeVersion);

		this.droidConf 					= new File(this.eRoot.getChild(E_DROID_CONF).getAttribute(A_PATH).getValue());
		this.jhoveConf 					= new File(this.eRoot.getChild(E_JHOVE_CONF).getAttribute(A_PATH).getValue());
		this.mPlayerPath 				= new File(this.eRoot.getChild(E_MPLAYER).getAttribute(A_PATH).getValue());
		this.mediaInfoPath 				= new File(this.eRoot.getChild(E_MEDIAINFO).getAttribute(A_PATH).getValue());
		this.iMagickIdentifyPath 		= new File(this.eRoot.getChild(E_IMAGICK_IDENTIFY).getAttribute(A_PATH).getValue());
		this.iMagickConvertPath 		= new File(this.eRoot.getChild(E_IMAGICK_CONVERT).getAttribute(A_PATH).getValue());

		this.defaultValidatorMessage 	= this.getXmlRootElement().getChild("defaultValidatorMessage").getValue();
		try{
		this.maxSizeMessage = Integer.parseInt(this.getXmlRootElement().getChild("maxSizeMessage").getValue());
		
		}catch(Exception e){
			this.maxSizeMessage=500;
		}
		
		// do on the applicatoin startup
		this.getFormatConfigList();
		this.getFormatConfigProdList();
		this.getFormatConfigTestList();
	}


	public String getHomeJavaProperty(){
		if(this.homeJavaProperty==null)	this.homeJavaProperty = System.getProperty("fvhome");
		return this.homeJavaProperty;
	}

	public File getHome(){
		if(this.home==null) this.home = new File(this.getHomeJavaProperty());
		return this.home;
	}

	public File getXmlFile(){
		if(this.xmlFile==null)
			this.xmlFile = new File(this.getHome(),"/etc/format-validator.xml");
		return this.xmlFile;
	}

	public File getXsdFile(){
		if(this.xsdFile==null)
			this.xsdFile = new File(this.getHome(),"/etc/format-validator.xsd");
		return this.xsdFile;
	}

	private Element getXmlRootElement(){
		if(this.eRoot==null)
			try {
				this.eRoot  = new SAXBuilder().build(this.getXmlFile()).getRootElement();
			} catch (JDOMException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			return this.eRoot;
	}
	
	public int getMaxSizeMessage(){
		return this.maxSizeMessage;
	}

	public static Config getInstance()   {
		if (instance == null) 	instance = new Config();
		return instance;
	}

	public Element getRootElement(){
		return this.eRoot;
	}

	public File getMPlayerPath(){
		return this.mPlayerPath;
	}

	public File getMediaInfoPath(){
		return this.mediaInfoPath;
	}

	public File getIMagickIdentifyPath(){
		return this.iMagickIdentifyPath;
	}

	public File getIMagickConvertPath(){
		return this.iMagickConvertPath;
	}

	public File  getLog4jConf(){
		if(this.log4jConf==null) this.log4jConf = new File(this.getXmlRootElement().getChild(Config.E_LOG4J_CONF).getAttribute(A_PATH).getValue());
		return  this.log4jConf;
	}

	public File  getDroidConf(){
		return  this.droidConf;
	}

	public String getConfigVersion(){
		if(this.version==null) this.version = this.getXmlRootElement().getChildText("version");
		return this.version;
	}

	public String getName(){
		if(this.name==null)	this.name = this.getXmlRootElement().getChildText("name");
		return this.name;
	}

	public File  getTmpDir(){
		if(this.tmpDir==null) this.tmpDir = new File(this.getXmlRootElement().getChild("tmp").getAttribute(A_PATH).getValue());
		return  this.tmpDir;
	}

	public File getJhoveConf(){
		return this.jhoveConf;
	}

	public String getDefaultValidatorMessage(){
		return this.defaultValidatorMessage;
	}


	public ArrayList<FormatConf>  getFormatConfigTestList() {
		if(this.formatConfigTestList!=null) return this.formatConfigTestList;
		this.formatConfigTestList = new ArrayList<FormatConf>();	
		for(FormatConf fc :  this.getFormatConfigList()){
			if(fc.getStatus().equalsIgnoreCase(FormatConf.STATUS_TEST))
				this.formatConfigTestList.add(fc);
		}
		return this.formatConfigTestList;
	}

	public ArrayList<FormatConf>  getFormatConfigProdList() {
		if(this.formatConfigProdList!=null) return this.formatConfigProdList;
		this.formatConfigProdList = new ArrayList<FormatConf>();	
		for(FormatConf fc :  this.getFormatConfigList()){
			if(!fc.getStatus().equalsIgnoreCase(FormatConf.STATUS_TEST))
				this.formatConfigProdList.add(fc);
		}
		return this.formatConfigProdList;
	}

	public ArrayList<FormatConf>  getFormatConfigList() {
		if(this.formatConfigList!=null) return this.formatConfigList;
		this.formatConfigList = new ArrayList<FormatConf>();	
		for(Object o :  this.getXmlRootElement().getChildren("format")){
			FormatConf fc = new FormatConf((Element)o);
			this.formatConfigList.add(fc);
		}
		return this.formatConfigList;
	}

	static void canRead(File file) throws Exception{
		if(!file.canRead()) throw new Exception("unable read file : "+file.getAbsolutePath());
	}
}