/**
 * @author prat@cines.fr
 * last maj : 2010/08/02
 * */

package fr.cines.pac.format;

public class Format {

	/**
	 * identifiant spécifiant qu'un valeur est dans l'etat "non applicatble"
	 * */
	public final static String NON_APPLICABLE = "NA";
	
	/**
	 * seprateur pour les encodages 1 et 2.
	 * */
	public final static String ENCODING_SEPARATOR = "/";
	
	private String 
	fileType, 		// type de fichier
	shortName, 		// nom court du format
	encoding,		// encodage du format : designe l'ensemble de l'encodage 1 et de l'encodage 2.
	encoding1,		// encodage numéro 1 du format : désigne par exemple le flux video pour une video
	encoding2,		// encodage numéro 1 du format : désigne par exemple le flux video pour une video
	version;		// designe la version du format
	
	public Format(){
		this.fileType 	= NON_APPLICABLE;
		this.shortName 	= NON_APPLICABLE;
		this.encoding 	= NON_APPLICABLE;
		this.encoding1 	= NON_APPLICABLE;
		this.encoding2 	= NON_APPLICABLE;
	}
	
	public Format(String shortName, String encoding){
		this();
		this.shortName = shortName;
		this.encoding = encoding;
		this.fileType =  shortName+ENCODING_SEPARATOR+encoding;
	}

	public String getFileType() {
		return fileType;
	}
	
	public void setFileType(String fileType) {
		this.fileType =  fileType;
		try{
			java.util.StringTokenizer tokenizer = new java.util.StringTokenizer(fileType, "/");
			this.shortName = tokenizer.nextToken();
			this.encoding = tokenizer.nextToken();
		}catch(Exception e){
			//do nothing
			//e.printStackTrace();
		}
	}
	public String getShortName() {
		return shortName;
	}

	public void setShortName(String shortName) {
		this.shortName = shortName;
	}

	public String getEncoding() {
		return encoding;
	}
	
	public boolean isEncodingExist(){	
		return !this.getEncoding().equalsIgnoreCase(NON_APPLICABLE) &&
		!this.getEncoding().isEmpty();
	}
	
	public void setEncoding(String encoding) {
		this.encoding = encoding;
	}

	public String getEncoding1() {
		return encoding1;
	}

	public void setEncoding1(String encoding1) {
		this.encoding1 = encoding1;
	}

	public String getEncoding2() {
		return encoding2;
	}

	public void setEncoding2(String encoding2) {
		this.encoding2 = encoding2;
	}
	
	public void setVersion(String version){
		this.version = version;
	}
	
	public String getVersion(){
		return this.version;
	}
}