/**
 * @author prat@cines.fr
 * last maj : 2010/08/02
 * */

package fr.cines.pac.format;
/**
 */

import java.util.List;

import fr.cines.pac.formatValidator.exception.FormatValidatorException;
import fr.cines.pac.formatValidator.exception.ValidatorNotFoundException;
import fr.cines.pac.formatValidator.util.MyElement;

import org.jdom.*;

/**
 * Classe de configuration d'un format à  partir d'un noeud XML designant un format 
 * */

public class FormatConf {

	/** nom de la balise xml designant le nom court du format*/
	public final static String E_shortName 			="shortName";
	
	/** nom de la balise xml designant le status du format (production ou test)*/
	public final static String E_status	 			="status";
	
	/** nom de la balise xml designant l'encodage du format*/
	public final static String E_pacEncoding 		="pacEncoding";

	/** nom de la balise xml designant la classe java de validation*/
	public final static String E_validatorClassName ="validatorClassName";
	
	/** nom de la balise xml designant le nom long du validateur*/
	public final static String E_validatorLongName  ="validatorLongName";
	
	/** nom de la balise xml designant la note ou description du validateur*/
	public final static String E_validatorNote		="validatorNote";
	
	/** nom de la balise xml designant une url vers une description web du validateur*/
	public final static String E_validatorUrl       ="validatorUrl";

	/** 
	 * nom de la balise xml designant les extentions de fichiers correspondante au format 
	 * ceci est utilisé pour l'execution en ligne de commande
	 * */
	public final static String E_extensions			="extensions";
	
	/** nom de la balise complexe xml designant les valeuts permises pour les encodages et nom de fichier */
	public final static String E_allow 				="allow";
	
	/** nom de la balise complexe xml designant les outils utilisés par le validateur */
	public final static String E_tools 				="tools";
	
	/** nom de la balise xml designant le nom long du format*/
	public final static String E_longname 			="longName";
	
	/** nom de la balise xml designant une note sur les format */
	public final static String E_note 				="note";

	/** nom de la balise xml designant le statut test*/
	public final static String STATUS_TEST			="test"; 

	private String 
	shortName, 
	status,
	pacEncoding,
	longName,
	validatorClassName,
	validatorLongName,
	validatorNote,
	validatorUrl,
	note;

	private String[]  extensions; 
	private FormatAllow allow;
	private Element rootElement;
	private List<Element> eTools;

	@SuppressWarnings("unchecked")
	public FormatConf(Element elRoot){
		this.setElRoot(elRoot);

		try{
			this.allow = new FormatAllow(this.getRootElement().getChild(E_allow));
		}catch(Exception ex){
			this.allow = new FormatAllow();
		}

		this.shortName 				= this.getERootChildValue(E_shortName);
		this.status 				= this.getERootChildValue(E_status);
		this.pacEncoding 			= this.getERootChildValue(E_pacEncoding);

		this.validatorClassName 	= this.getERootChildValue(E_validatorClassName);
		this.validatorLongName		= this.getERootChildValue(E_validatorLongName);
		this.validatorNote			= this.getERootChildValue(E_validatorNote);
		this.validatorUrl 			= this.getERootChildValue(E_validatorUrl);

		this.eTools 	= this.getRootElement().getChildren(E_tools);
		this.longName 	= this.getERootChildValue(E_longname);
		this.note 		= this.getERootChildValue(E_note);

		String[] tmp =this.getERootChildValue(E_extensions).split(",", 0);
		int i=0;
		this.extensions = new String[tmp.length];
		for(String s : tmp){
			this.extensions[i] = s.trim();
			i++;
		}
	}

	public static boolean isFileNameSuffix(String fileName, String[] suffixArray ){
		fileName = fileName.toLowerCase();
		for(String suffix : suffixArray)
			if(fileName.endsWith(suffix.toLowerCase())) return true;
		return false;
	}

	public List<Element> getEtools(){
		return this.eTools;
	}

	private void  setElRoot(Element rootElement){
		this.rootElement = rootElement;
	}

	public Element getRootElement(){
		return this.rootElement;
	}

	public String getERootChildValue(String childName){
		return MyElement.getValue(this.getRootElement().getChild(childName));
	}

	public String getStatus(){
		if(this.status==null) return "";
		return this.status;
	}

	public boolean isInProduction(){
		if(this.getStatus()==null) return true;
		return this.getStatus().isEmpty();
	}

	public String getPacEncoding(){
		return this.pacEncoding;
	}

	/**
	 * @return the shortName
	 */
	public String getshortName() {
		return this.shortName;
	}

	public String getLongName() {
		return this.longName;
	}

	public String getNote() {
		return this.note;
	}

	public String getValidatorLongName() {
		return validatorLongName;
	}

	public String getValidatorNote() {
		return validatorNote;
	}

	public String getValidatorUrl() {
		return validatorUrl;
	}

	/**
	 * @return the suffixFileName
	 */
	public String[] getExtensions() {
		return this.extensions;
	}

	public static FormatConf getFromShortName(String idPacFormat) throws FormatValidatorException {
		return getFromAttrs(idPacFormat, null);
	}
	
	public static FormatConf getFromAttrs(String idPacFormat, String encoding) throws FormatValidatorException {
		if(encoding!=null){
			for(FormatConf f :  Config.getInstance().getFormatConfigList())
				if(f.getshortName().equalsIgnoreCase(idPacFormat)
						&&
						f.getPacEncoding().equalsIgnoreCase(encoding)	
				) return f;
		}
		for(FormatConf f :  Config.getInstance().getFormatConfigList())
			if(f.getshortName().equalsIgnoreCase(idPacFormat)) return f;
		
		throw new ValidatorNotFoundException("unable to load Validator from shortName=" +idPacFormat);
	}

	public static FormatConf getFormatConfFromFileName(String fileName) throws FormatValidatorException {
		String targetExtension = getSuffix(fileName).toLowerCase();
		for(FormatConf f :  Config.getInstance().getFormatConfigList() )
			for(String currentExtension : f.getExtensions())
				if(currentExtension.equalsIgnoreCase(targetExtension))	return f;
		return null;
	}

	public String getValidatorClassName(){
		return this.validatorClassName;
	}

	public FormatAllow getAllow(){
		return this.allow;
	}

	public static String getSuffix(String in){
		return in.substring(in.lastIndexOf(".")+1, in.length());
	}

	public static FormatConf getFormatConfigFromValidator(String className) throws FormatValidatorException{
		for(FormatConf fc : Config.getInstance().getFormatConfigList()){
			if(fc.getValidatorClassName().compareTo(className)==0) return fc;
		}
		return null;
	}
}