/**
 * @author prat@cines.fr
 * last maj : 2010/08/02
 * */
 
package fr.cines.pac.formatValidator;

import fr.cines.pac.formatValidator.exception.*;
import fr.cines.pac.formatValidator.tools.*;

/**
 * Classe permettant la validation des fichiers au format MPEG4
 */

public class AudioValidator extends  Validator {

	@Override
	void setFormatToolsInfo() {
		this.getFormatToolsInfo().setShortName(this.getToolsMgr().getMediaInfoTool().getGeneralFormat());
		this.getFormatToolsInfo().setEncoding1(this.getToolsMgr().getMediaInfoTool().getAudioFormat());	
	}
	
	@Override
	void checkIsWellFormed() throws NotWellFormedException {
		this.checkAllowFormat();
		this.checkAllowAudioEncoding();
		this.checkAudioIsOnly();
		this.checkAudioIsOnlyOne();
	}
	

	@Override
	void checkIsValid() throws NotValidException {
		this.getToolsMgr().getMPlayerAudioTool().run();
		
	//	System.out.println("exitvalue="+this.getToolsMgr().getMPlayerAudioTool().getProcessTool().getInput());
	//	System.out.println("exitvalue="+this.getToolsMgr().getMPlayerAudioTool().getProcessTool().getExitValue());
		if(!this.getToolsMgr().getMPlayerAudioTool().getProcessTool().getInput().isEmpty())
			throw new NotValidException(this.getClass(),this.getToolsMgr().getMPlayerAudioTool().getProcessTool().getInput());
		//if(!this.getToolsMgr().getMPlayerAudioTool().getProcessTool().getInput().contains(
			//	"Do not report crashes to FFmpeg developers."));
			/*	
		if(this.getToolsMgr().getMPlayerAudioTool().getProcessTool().getExitValue()!=0)
			throw new NotValidException(this.getClass(),this.getToolsMgr().getMPlayerAudioTool().getProcessTool().getInput());
	*/}

	private void checkAllowAudioEncoding() throws NotWellFormedException{
		if(!this.getFormatConf().getAllow().getEncoding1().contains(this.getFormatToolsInfo().getEncoding1()))
			throw new NotWellFormedException(this.getClass(),"les encodages audio admis sont : "+this.getFormatConf().getAllow().getEncoding1().toString()+" or l'encodage trouvé est : "+this.getFormatToolsInfo().getEncoding1());
	}

	private void checkAudioIsOnly() throws NotWellFormedException{
		String videoFormatList = this.getToolsMgr().getMediaInfoTool().getGeneral(MediaInfoTool.GENERAL_VIDEO_FORMAT_LIST);
		if(!videoFormatList.equalsIgnoreCase(MediaInfoTool.NOT_FOUND)){
			throw new NotWellFormedException(this.getToolsMgr().getMediaInfoTool().getClass(),"le fichier ne doit pas contenir le flux video "+videoFormatList);
		}
		
		if(!this.getToolsMgr().getMediaInfoTool().getVideoFormat().equalsIgnoreCase(MediaInfoTool.NOT_FOUND))
			throw new NotWellFormedException(this.getToolsMgr().getMediaInfoTool().getClass(),"le fichier ne doit pas contenir le flux video : "+this.getToolsMgr().getMediaInfoTool().getVideoFormat());
	}
	
	private void checkAudioIsOnlyOne() throws NotWellFormedException{
		String audioStreamCount = this.getToolsMgr().getMediaInfoTool().getReader().getGeneral().get(MediaInfoTool.GENERAL_COUNT_OF_AUDIO_STREAM);
		if(!audioStreamCount.equalsIgnoreCase("1")){
			throw new NotWellFormedException(this.getToolsMgr().getMediaInfoTool().getClass(),"le fichier ne doit pas contenir plus d'un flux audio. nombre de flux trouvé :  "+audioStreamCount);
		}
	}
}