/**
 * @author prat@cines.fr
 * last maj : 2010/08/02
 * */

package fr.cines.pac.formatValidator;

import java.io.File;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.jdom.Element;
import org.jdom.input.SAXBuilder;
import org.w3c.dom.Document;
import edu.harvard.hul.ois.jhove.module.XmlModule;
import fr.cines.pac.format.*;
import fr.cines.pac.formatValidator.exception.*;


/**
 * Classe permettant la validation des fichiers au format SVG
 */
public class SvgValidator extends Validator {
	
	public static final int[] validDroidIds = {634,635};

	public void init(File file) throws FormatValidatorException{
		super.init(file);
		this.getToolsMgr().setJhoveTool(new XmlModule());
		this.getToolsMgr().setDroidTool(validDroidIds);
	}
	

	
	public String getVersion(){
		String version = Format.NON_APPLICABLE;
		try {
			Element eRoot  = new SAXBuilder().build(this.getFile()).getRootElement();
			version = eRoot.getAttribute("version").getValue();
			System.out.println(eRoot.getName());
		} catch (Exception e) {
			version = Format.NON_APPLICABLE;
		} 		
		return version;
	}
	
	
	@Override
	void checkIsValid() throws NotValidException {
		this.getToolsMgr().getIMagickTool().checkIsValid();
		this.getToolsMgr().getJhoveTool().checkIsValid();
		this.getToolsMgr().getDroidTool().checkIsValid();
		//this.getToolsMgr().getIMagickTool().checkIsValid();
		this.getToolsMgr().getCinesXmlTool().checkIsValid();
	}

	@Override
	void checkIsWellFormed() throws NotWellFormedException {
		this.getToolsMgr().getIMagickTool().checkIsWellFormed("SVG");
		// on prend toutes les version this.checkAllowVersion();
		this.getToolsMgr().getJhoveTool().checkIsWellFormed();
		this.getToolsMgr().getCinesXmlTool().checkIsWellFormed();
		
		this.getToolsMgr().getDroidTool().checkIsWellFormed();
	//	this.droid.init(this.getFile());
		
	}

	@Override
	void setFormatToolsInfo() {
		this.getFormatToolsInfo().setShortName(this.getToolsMgr().getIMagickTool().getShortName());		
		// pas mieux pour le moment
		this.getFormatToolsInfo().setVersion(this.getVersion());
	}
	
	private String  getDoctypeVersion_() {
		// suppression dû à un mauvais résultat retourné par droid ; à investiguer
		DocumentBuilderFactory fabrique = DocumentBuilderFactory.newInstance();
		try{
			DocumentBuilder documentBuilde = fabrique.newDocumentBuilder();
			Document document = documentBuilde.parse(this.getFile());
			if (document.getDoctype() == null) return Format.NON_APPLICABLE;
			if (document.getDoctype().getPublicId() == null) return Format.NON_APPLICABLE;
			
			String publicId = document.getDoctype().getPublicId();
			Pattern p = Pattern.compile("-//W3C//DTD SVG ([0-9.]*)//EN");
			Matcher m = p.matcher(publicId);
			          	   
			if (m.find()) return m.group(1);
		}
		catch (Exception e)	{
			e.printStackTrace();
			// do nothing this.setMessage(e.getLocalizedMessage());
		}
		return Format.NON_APPLICABLE;		
	}
}
