

package fr.cines.pac.formatValidator;


import java.io.File;
import java.util.ArrayList;

import org.apache.log4j.Logger;
import org.apache.log4j.helpers.Loader;

import fr.cines.pac.format.*;
import fr.cines.pac.formatValidator.exception.*;
import fr.cines.pac.formatValidator.tools.ToolsMgr;
import fr.cines.pac.formatValidator.util.Util;


/**
 * Méthode d'initialisation devant impérativement être appelée pour pouvoir utiliser
 * la classe.
 * @param file Fichier
 * @throws ArcsysCinesException
 */
public abstract class  Validator implements FileFormatValidator {

	// affectation du nom du logger
	Logger logger = Logger.getLogger(this.getClass().getName());

	// message donné par le validateur
	String 	message="";

	//configuration du format voir fichiers de configuration dans repertoire etc
	FormatConf formatConf;

	//format trouvé par les outils de controles.
	Format formatToolsInfo;

	// fichier analysé
	private File file;

	/**
	 * 	caractèrise le respect de la sémentique du format.
	 *   0 le test n'est pas réalisé
	 *   1 test ok
	 *  -1 test ko
	 */
	private int  isValid;
	
	
	/**
	 * 	caractèrise le respect de la syntaxe du format.
	 *   0 le test n'est pas réalisé
	 *   1 test ok
	 *  -1 test ko
	 */
	private int  isWellFormed;

	boolean formatOk, encoding1Ok, encoding2Ok;

	ToolsMgr toolsMgr;

	public Validator(){
	}

	@Deprecated
	public void  initFile(File file, String doNothing) throws FormatValidatorException{
		initFile(file);
	}
	@Deprecated
	public void  initFile(File file) throws FormatValidatorException{
		init(file);
	}

	public void  init(File file) throws FormatValidatorException{
		this.isValid = 0;
		this.isWellFormed = 0;
		this.message="";
				
		if(!Config.getInstance().isCheck()) Config.getInstance().check();

		//initialisation des logs
		logger.debug("init : "+this.getClass().getSimpleName());

		this.file = file;
		if(!this.getFile().canRead())
			throw new FormatValidatorException("could not read file : "+this.getFile().getAbsolutePath());	
	}

	public FormatConf getFormatConf(){
		if(this.formatConf==null){
			ArrayList<FormatConf> formatConfigureList= Config.getInstance().getFormatConfigList();
			for(FormatConf fc : formatConfigureList){
				if(fc.getValidatorClassName().equalsIgnoreCase(this.getClass().getSimpleName())){
					this.formatConf = fc;
					return this.formatConf;
					//break;
				}
			}
		}
		return this.formatConf;
	}


	public ToolsMgr getToolsMgr(){
		if(this.toolsMgr==null) this.toolsMgr= new ToolsMgr(this.getFile());
		return this.toolsMgr;
	}

	//abstract void initValidation() throws NotWellFormedException, NotValidException, FormatValidatorException;

	abstract void checkIsWellFormed() throws NotWellFormedException;

	abstract void checkIsValid() throws NotValidException;

	abstract void setFormatToolsInfo();

	public  Format getFormatToolsInfo() {
		if(this.formatToolsInfo==null){
			this.formatToolsInfo = new Format();
			this.setFormatToolsInfo();
		}
		return this.formatToolsInfo;
	}

	public String getVersionFormat(){
		return this.getFormatToolsInfo().getVersion();
	}

	void checkAllowFormat() throws NotWellFormedException{
		if(!this.getFormatConf().getAllow().getFormat().contains(this.getFormatToolsInfo().getShortName()))
			throw new NotWellFormedException(this.getClass(), "les format admis sont : "+this.getFormatConf().getAllow().getFormat().toString()+", le format trouvé est : "+this.getFormatToolsInfo().getShortName() );
	}

	void checkAllowEncoding1() throws NotWellFormedException{
		if(!this.getFormatConf().getAllow().getEncoding1().contains(this.getFormatToolsInfo().getEncoding1()))
			throw new NotWellFormedException(this.getClass(),"les encodages de type 1 admis sont : "+this.getFormatConf().getAllow().getEncoding1().toString()+" or l'encodage trouvé est : "+this.getFormatToolsInfo().getEncoding1());
	}

	void checkAllowEncoding2() throws NotWellFormedException{
		if(!this.getFormatConf().getAllow().getEncoding2().contains(this.getFormatToolsInfo().getEncoding2()))
			throw new NotWellFormedException(this.getClass(),"les encodages de type 2 admis sont  : "+this.getFormatConf().getAllow().getEncoding2().toString()+" or l'encodage trouvé est : "+this.getFormatToolsInfo().getEncoding2());
	}


	void checkAllowVersion() throws NotWellFormedException{
		if(!this.versionMustBeChecked()) return;
		if(!this.getFormatConf().getAllow().getVersion().contains(this.getFormatToolsInfo().getVersion())){
			throw new NotWellFormedException(this.getClass(),"les versions  admises sont : "+this.getFormatConf().getAllow().getVersion().toString()+", la verion trouvée est "+this.getFormatToolsInfo().getVersion());
		}
	}

	boolean versionMustBeChecked(){
		if(this.getFormatConf().getAllow().getVersion()==null) return false;
		if(this.getFormatConf().getAllow().getVersion().isEmpty()) return false;
		return true;
	}

	public boolean isArchivable(){
		return this.isValid();
	}

	/**
	 * indique si le format est valide
	 */
	public  boolean isValid(){
		if(!this.isWellFormed())return false;

		if(this.isValid==0){
			try{
				this.checkIsValid();
				this.isValid=1;
			}
			catch(NotValidException e){
				this.isValid = -1;
				this.setMessage(e.getToolClass().getSimpleName()+" message : "+e.getLocalizedMessage());
			}
		}
		return this.isValid==1;
	}

	/**
	 * indique si le format est bien formé
	 */
	public  boolean isWellFormed(){
		if(this.isWellFormed==0){
			try{
				this.checkIsWellFormed();
				this.isWellFormed=1;
			}
			catch(NotWellFormedException e){
				this.isWellFormed = -1;
				this.setMessage(e.getToolClass().getSimpleName()+" message : "+e.getLocalizedMessage());
			}
		}
		return this.isWellFormed==1;
	}

	public File getFile(){
		return this.file;
	}

	/**
	 * @return message object
	 * @throws FormatValidatorException 
	 */
	public   String getMessage(){
		int maxSizeMessage = Config.getInstance().getMaxSizeMessage();
		
		String prefix =  Config.getInstance().getName()+" "+
		
		Config.codeVersion+ " ";
		
		String defaultMessage = prefix+", "+Config.getInstance().getDefaultValidatorMessage();
		
		if(this.message==null)	return defaultMessage;
		if(this.message.isEmpty()) return defaultMessage;
		
		if(this.message.length()>maxSizeMessage) 
			try{
				this.message = this.message.substring(0, maxSizeMessage);
				this.message = this.message+"..... ( message tronqué, caractères limités à "+maxSizeMessage+" ) ";
			}catch(Exception e){
				// do nothing
			}
		
		return prefix+", "+this.message;
	}

	private void setMessage(String message){
		this.message = message;
	}
	
	/**
	 * @return information sur la validation du fichier
	 */
	public String getSortInfo(){ 
		return  
		"bien formé="+Util.unzero(this.isWellFormed())+
		", valide="+Util.unzero(this.isValid())+
		", archivable="+Util.unzero(this.isArchivable())+
		", fichier["+file.getAbsolutePath()+"]"+
		", message : "+this.getMessage();
	}

	String getId(){
		return this.getClass().getSimpleName();
	}

	public static Validator getValidatorFromPacFormatId(String shortName) throws FormatValidatorException  {
		String validatorClassName ="NaValidator";
		if(shortName!=null)
			validatorClassName = FormatConf.getFromShortName(shortName).getValidatorClassName();
		return getFromClassName(validatorClassName);
	}

	public static Validator getValidatorFromPacFormatId(String shortName, String encoding) throws FormatValidatorException  {
		String validatorClassName = FormatConf.getFromAttrs(shortName, encoding).getValidatorClassName();
		return getFromClassName(validatorClassName);
	}

	public static Validator getFromClassName(String validatorClassName) throws ValidatorNotFoundException  {
		String fullClassName = "fr.cines.pac.formatValidator."+validatorClassName;
		try {
			return  (Validator)Loader.loadClass(fullClassName).newInstance();
		} catch (Exception e) {
			throw new ValidatorNotFoundException("error loading validator class : "+fullClassName,e);
		} 
	}

}