/**
 * @author prat@cines.fr
 * last maj : 2010/08/02
 * */

package fr.cines.pac.formatValidator.tools;

import java.io.File;

import uk.gov.nationalarchives.droid.AnalysisController;
import uk.gov.nationalarchives.droid.IdentificationFile;
import uk.gov.nationalarchives.droid.signatureFile.FileFormat;
import fr.cines.pac.format.Format;
import fr.cines.pac.formatValidator.exception.NotValidException;
import fr.cines.pac.formatValidator.exception.NotWellFormedException;

/**
 * implémentation de l'outil de validation Droid
 */

public class DroidTool extends Tool{

	private DroidToolSingleton droidSingleton;
	
	private IdentificationFile idf;
	private int validIds[];	
	
	
	public DroidTool(int[] validIds) {
		this.validIds = validIds;
		droidSingleton = DroidToolSingleton.getInstance();
	}
	
	public void init(File file){
		logger.debug("init DroidTool with validids="+this.getvalidIdsString());
		idf = droidSingleton.getIdentification(file.getAbsolutePath());
	}

	private boolean isValidId(int id) {
		for(int j : validIds){
			if(j == id) return true;
		}
		return false;
	}
	
	private FileFormat gotOneValidFormat() {
		for (int i = 0; i < idf.getNumHits(); i++) {
			if (isValidId(idf.getHit(i).getFileFormat().getID())) return idf.getHit(i).getFileFormat();
		}
		return null;
	}
	
	private boolean isPositive() {
		return (idf.getClassification()==AnalysisController.FILE_CLASSIFICATION_POSITIVE);
	}
	
	private boolean isTentative() {
		return (idf.getClassification()==AnalysisController.FILE_CLASSIFICATION_TENTATIVE);
	}
	
	public String getFileFormat() {
		FileFormat f = gotOneValidFormat();
		if (f!=null) return f.getName();
		else {
			if(idf.getNumHits() > 0) 
			{
				logger.debug(idf.getHit(0).getFileFormatVersion());
				return idf.getHit(0).getFileFormatName();
			}
			else return null;
		}
	}

	public String getVersionFormat() {
		FileFormat f = this.gotOneValidFormat();
		if (f!=null) return f.getVersion();
		else {
			if(idf.getNumHits() > 0) return idf.getHit(0).getFileFormatVersion();
			else return null;
		}
	}

	private boolean isValid() {
		if((this.gotOneValidFormat() == null) || (!this.isPositive())) return false;
		else return true;
	}

	private boolean isWellFormed() {
		if((gotOneValidFormat() == null) || (!isPositive() && !isTentative())) return false;
		else return true;
	}

	public void checkIsValid() throws NotValidException {
		if(!this.isValid()) throw new NotValidException(this.getClass(), "l'analyse es negative."+"  "+idf.getText());
	}

	public void checkIsWellFormed() throws NotWellFormedException {
		if(!this.isWellFormed()) throw new NotWellFormedException(this.getClass(), "echec de la tentative d'analyse."+"  "+idf.getText());
	}
	
	private String getvalidIdsString(){
		String out="";
		for(int j : validIds){
			out = out+j+" ";
		}
		return out;
	}
	
	public String getFormatPacId(){
		String in = this.getFileFormat();
		if(in==null) return Format.NON_APPLICABLE;
		if(in.equalsIgnoreCase("Portable Network Graphics")) return "PNG";
		if(in.equalsIgnoreCase("Scalable Vector Graphics")) return "SVG";
		return in;
	}
}