package fr.cines.pac.formatValidator.tools;

import java.io.File;
import java.util.List;

import fr.cines.pac.format.Format;
import fr.cines.pac.formatValidator.exception.*;

import edu.harvard.hul.ois.jhove.*;

public class JhoveTool extends Tool{

	private RepInfo info;
	private String message;

	/**
	* implémentation de l'outil de validation jhove
	*/
	
	public JhoveTool(ModuleBase module, File file){
		info = new RepInfo(file.getAbsolutePath());
		boolean rval = false;
		try{
			rval = new JhoveToolSingleton().processFile(module, false, file, info);
		}catch(Exception e){
			this.message = e.getMessage();
		}
		if (!rval) {
			info.setValid(RepInfo.FALSE);
			info.setWellFormed(RepInfo.FALSE);
		}
		System.gc();
	}


	public String getFileFormat() {
		if(info == null) return null;
		return info.getFormat();
	}

	public String getVersionFormat() {
		if(info == null)  return Format.NON_APPLICABLE;
		if(info.getVersion()==null) return Format.NON_APPLICABLE;
		return info.getVersion();
	}

	public boolean isWellFormed() {
		if(info == null)  return false;
		return (info.getWellFormed() != RepInfo.FALSE);   
	}

	public  void checkIsWellFormed() throws NotWellFormedException {
		if(!this.isWellFormed()) throw new NotWellFormedException(this.getClass(),this.getMessage());
	}

	public  void checkIsValid() throws NotValidException {
		if(!this.isValid()) throw new NotValidException(this.getClass(), this.getMessage());
	}

	public  void checkProfil(String declaratedProfil) throws NotWellFormedException {
		String profil= this.getProfil(0);
		if(profil==null){
			throw new NotWellFormedException(this.getClass(), "impossible de trouver le profil"+ "  "+this.getMessage());
		}    	
		if(!profil.equals(declaratedProfil)){
			throw new NotWellFormedException(this.getClass(), "le profil déclaré est "+declaratedProfil+ ", alors que le profile trouvé est "+profil+"  "+this.getMessage());
		}
	}

	public boolean isValid() {
		if(info == null) return false;
		return (info.getValid() != RepInfo.FALSE);      
	}

	@SuppressWarnings("unchecked")
	public String getMessage(){
		if(message!=null ) return message;

		if( info==null) {
			this.message =  "no information from jhove";
			return this.message;
		}

		StringBuffer sb = new StringBuffer();
		List<Message> list = info.getMessage();
		for(Message m : list){
			String subM = m.getSubMessage();
			sb.append(m.getMessage());
			if(subM!=null)
				if(subM.length()>0)
					sb.append(", "+subM+" ");
		}
		this.message = sb.toString();

		return this.message;
	}

	@SuppressWarnings("unchecked")
	private List<String> getProfil()    {
		if(info == null) return null;
		List<String> profil = info.getProfile();
		if (profil == null) return null;
		else return profil;
	}

	public String getProfil(int index){
		try{
			return this.getProfil().get(index);
		}catch(Exception e){
			return null;
		}
	}
}