package fr.cines.pac.formatValidator.tools;

import java.io.File;
import java.util.ArrayList;

import fr.cines.pac.format.Config;
import fr.cines.pac.formatValidator.exception.NotValidException;

/**
 * Classe permettant la validation de fichiers videos en utilisant MPlayer's Movie Encoder
 * @see  http://www.mplayerhq.hu/DOCS/HTML/en/MPlayer.html#howtoread
 * @see : http://www.mplayerhq.hu/DOCS/HTML/en/mencoder.html
 */
public class MEncoderTool extends Tool {

	private ArrayList<String> authorizedAudioErrors = new ArrayList<String>();
	private ArrayList<String> authorizedVideoErrors = new ArrayList<String>();
	private ArrayList<String> authorizedContainerErrors = new ArrayList<String>();

	// commande posix pour l'outil mencoder
	// @see : http://www.mplayerhq.hu/DOCS/HTML/en/mencoder.html
	
	private String message = "";//Erreur indéterminée : tester le fichier en lecture.";

	private String specAudio = null;
	private String specVideo = null;
	private ProcessHandler processTool;	
	
	public MEncoderTool(File file){
		this.file = file;
		//authorizedAudioErrors.add(" FAAD: compressed input bitrate missing, assuming 128kbit/s!");
	}
	
	public static String[] getCmd(File file){
		String cmd[]= new String[13];
		String root = Config.getInstance().getMPlayerPath().getAbsolutePath();
		//	String command[]= new String[13];
		cmd[0]=root+"/bin/mencoder";
		cmd[1]="-ovc";  // output video codec [mencoder -ovc help]
		cmd[2]="raw";  	//  uncompressed video.
		cmd[3]="-oac"; 	// output audio codec [mencoder -oac help]
		cmd[4]="pcm";	// uncompressed PCM audio
		cmd[5]="-o";	// output file
		cmd[6]="/dev/null"; // output file path
		cmd[7]="-msglevel"; //niveau du message
		cmd[8]="all=2";		//?
		cmd[9]="-msgmodule";//?
		cmd[10]="-msglevel";//?
		cmd[11]="statusline=5";//?
			
		cmd[12]=file.getAbsolutePath();
		return cmd;
	}

	private ProcessHandler getProcessTool(){
		if(this.processTool==null){
			this.processTool = new ProcessHandler(MEncoderTool.getCmd(this.getFile()));
			processTool.run();
		}
		return processTool;
	}
	
	private boolean getStatus(String out) {
		int j;
		int i = out.indexOf("STATUSLINE");
		while (i>0) {
			out=out.substring(i+10);
			j = out.indexOf("STATUSLINE");
			if (j<0) break;
			i=j;
		}
		if (i>0) {
			out = out.substring(23, 26);
			//out=" 8%";
			if (out.charAt(2)=='%') out = out.substring(0, 2);
			if (out.charAt(0)==' ') out=out.substring(1);

			int status = Integer.parseInt(out);
			if (status>95) return true;
			else message="décodage arrété prématurément à : " + status + '%';
		}

		return false;
	}

	private boolean checkSpecAudio(String out) {
		if (specAudio == null) return true;
		else {
			StringBuffer audio = new StringBuffer();
			readInfo(out, specAudio, audio);
			if (audio.toString().equals("")) { return true; }
			else {
				message+=audio.toString()+'\n'; 
				return false; }
		}
	}

	private boolean checkSpecVideo(String out) {
		if (specVideo == null) return true;
		else {
			StringBuffer video = new StringBuffer();
			readInfo(out, specVideo, video);
			if (video.toString().equals("")) { return true; }
			else {
				message+=video.toString()+'\n'; 
				return false; }
		}
	}

	private boolean checkAudio(String out) {
		StringBuffer audio = new StringBuffer();
		int ind = readInfo(out, "DECAUDIO:", audio);
		if (audio.toString().equals("")) { return true; }
		else {
			if (!authorizedAudioErrors.contains(audio.toString())) { message+=audio.toString()+'\n'; return false; }
			else return checkAudio(out.substring(ind+audio.length()));
		}
	}

	private boolean checkContainer(String out) {
		StringBuffer demux = new StringBuffer();
		int ind = readInfo(out, "DEMUX:", demux);
		if (demux.toString().equals("")) { return true; }
		else {
			if (!authorizedContainerErrors.contains(demux.toString())) { message+=demux.toString()+'\n'; return false; }
			else return checkContainer(out.substring(ind+demux.length()));
		}
	}

	private boolean checkVideo(String out) {
		StringBuffer video = new StringBuffer();
		int ind = readInfo(out, "DECVIDEO:", video);
		if (video.toString().equals("")) { return true; }
		else {
			if (!authorizedVideoErrors.contains(video.toString())) { message+=video.toString()+'\n'; return false; }
			else return checkVideo(out.substring(ind+video.length()));
		}
	}


	private  boolean isValid() {
	//	if(this.getProcessTool().getProcess().exitValue()!=0) return false;
		boolean checkAudio = checkAudio(this.getProcessTool().getInput());
		boolean checkVideo = checkVideo(this.getProcessTool().getInput());
		boolean checkContainer = checkContainer(this.getProcessTool().getInput());
		boolean checkSpecAudio = checkSpecAudio(this.getProcessTool().getInput());
		boolean checkSpecVideo = checkSpecVideo(this.getProcessTool().getInput());
		boolean getStatus = getStatus(this.getProcessTool().getInput());
		boolean isValid = (checkAudio && checkVideo && checkContainer  && checkSpecAudio && checkSpecVideo && getStatus);
		return isValid;
	}
	
	public void checkIsValid() throws NotValidException{
		if(!this.isValid()){
			throw new NotValidException(this.getClass(), this.message);
		}
	}

	private static int readInfo(String out, String id, StringBuffer info) {
		int i = out.indexOf(id);
		int j=i;
		if (i>=0) {
			//i+=id.length();
			while (out.charAt(i) != '\n') {
				info.append(out.charAt(i));
				i++;
			}
		}
		return j;
	}

	public void addAuthorizedAudioErrors(String error) {
		authorizedAudioErrors.add(error);
	}

	public void addAuthorizedVideoErrors(String error) {
		authorizedVideoErrors.add(error);
	}

	public void addAuthorizedContainerErrors(String error) {
		authorizedContainerErrors.add(error);
	}

	public void setSpecAudio(String a) {
		this.specAudio=a;
	}

	public void setSpecVideo(String v) {
		this.specVideo=v;
	}

	public String getMessage() {
		return message;
	}
}