package fr.cines.pac.formatValidator.tools;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.OutputStream;
import java.io.PrintStream;
import java.util.StringTokenizer;

import fr.cines.pac.format.*;
import fr.cines.pac.formatValidator.exception.*;

/**
 * Classe permettant la validation de fichiers en utilisant ODFValidator
 *
 */
public class OdfTool  extends Tool{

	private OdfToolSingleton odf;
	private String mediaType;
	
	private String version = Format.NON_APPLICABLE;
	private boolean isValid;

	File file;

	String result;
	Exception odfInitException;

	public OdfTool(File file){
		this.file = file;
		try{
			this.odf = OdfToolSingleton.getInstance();
		}catch(Exception e){
			odfInitException = e;
		}
		OutputStream out = new ByteArrayOutputStream();

		this.isValid = odf.validate(file.getAbsolutePath(), new PrintStream(out));
		this.result = out.toString();
		this.setMediaType(this.result);
		this.setVersion(this.result);
	}

	public  void checkIsValid() throws NotValidException {
		if(!this.isValid) throw new NotValidException(this.getClass(), this.result);
		StringTokenizer st = new StringTokenizer(this.result, "\r\n");
		String message="";

		while (st.hasMoreTokens()) {
			String line = st.nextToken();
			if(line.indexOf(":Error:")>0)
				message += line.substring(file.getParentFile().getAbsolutePath().length()+1);
		}

		if(!message.isEmpty()) throw new NotValidException(this.getClass(),message);
	}


	private void setMediaType(String output){
		String flag = "Media Type:";
		int beginIndex = output.indexOf(flag)+flag.length();
		int endIndex = output.indexOf('\n',beginIndex);
		this.mediaType = output.substring(beginIndex, endIndex).trim();
	}

	private void  setVersion(String output){
		String flag = "ODF Version:";
		int beginIndex = output.indexOf(flag)+flag.length();
		int endIndex = output.indexOf('\n',beginIndex);
		this.version=  output.substring(beginIndex, endIndex).trim();
	}

	public String getVersion(){
		return this.version;
	}

	/*
	public boolean isValid() {
		boolean result;
		out = new ByteArrayOutputStream();
		PrintStream p = new PrintStream(out);
		result = ((odf.validateStrict(file.getAbsolutePath(), p)) && (getFileFormat().equals(mediaType)));
		message+="Validation stricte :\n"+out.toString()+'\n'; 
		return result;
	}


	public boolean isWellFormed() {
		boolean result;
		out = new ByteArrayOutputStream();
		PrintStream p = new PrintStream(out);
		result = ((odf.validate(file.getAbsolutePath(), p)) && (getFileFormat().equals(mediaType)));
		message+="Validation standard :\n"+out.toString()+'\n';
		return result;
	}
	*/

	public String getMediaType(){
		return this.mediaType;
	} 
}
