package fr.cines.pac.formatValidator.tools;

import java.io.*;

import javax.xml.parsers.*;
import javax.xml.transform.stream.StreamSource;
import javax.xml.validation.*;
import org.w3c.dom.Document;

import org.xml.sax.*;

import fr.cines.pac.format.Format;
import fr.cines.pac.formatValidator.exception.*;

/**
 * Classe validant les fichiers XML
 *
 */
public class SaxXmlTool  extends Tool{

	private File file;
	private String message ="";

	public SaxXmlTool(File file) {
		this.file = file;
	}

	public boolean isValid(){
		try{
			validerFichierXML(file.getAbsolutePath(), true);
		}
		catch (Exception e){
			this.message = e.getMessage();
			return false;
		}
		return true;
	}


	public String getMessage() {
		return this.message;
	}

	/**
	 * Valide un fichier XML de nom spécifié
	 * @param cheminFichier
	 * @param validerSchemaOuDtd Si ce paramètre vaut false, on fait uniquement une validation de premier
	 * niveau (XML well-formed). S'il vaut true on fait une validation avec le schéma ou la DTD spécifiés
	 * dans le fichier XML.
	 */

	public void validerFichierXML(String path, boolean validerSchemaOuDtd) throws FormatValidatorException
	{
		XmlSchemaTool validator = new XmlSchemaTool();
		try {
			InputSource inputSource = new InputSource(new FileInputStream(path));
			File xmlFileDirectory = new File(path).getParentFile();
			//      if (logger.isDebugEnabled())
			//       logger.debug("validerFichierXML : xmlFileDirectory=[" + xmlFileDirectory.toString() + "]");
			//Il faut faire cette manipulation pour que la résolution locale
			//de DTD fonctionne !
			inputSource.setSystemId(xmlFileDirectory.getAbsolutePath() + "/");
			//if (logger.isDebugEnabled())
			logger.debug("check xml   : validerSchemaOuDtd="+validerSchemaOuDtd+"   "+path+" inputSource.setSystemId=" + inputSource.getSystemId() );
			validator.validateXML(inputSource, validerSchemaOuDtd);
		}
		catch (IOException e) {
			throw new FormatValidatorException(e);
		}
		catch (SAXException e) {
			throw new FormatValidatorException(e.getMessage());
		}

	}

	public static void validerFichierXMLAvecXSD(String fichierxml, String fichierxsd) throws FormatValidatorException {
		try {
			// define the type of schema - we use W3C:
			String schemaLang = "http://www.w3.org/2001/XMLSchema";

			// get validation driver:
			SchemaFactory factory = SchemaFactory.newInstance(schemaLang);

			// create schema by reading it from an XSD file:
			Schema schema = factory.newSchema(new StreamSource(fichierxsd));
			Validator validator = schema.newValidator();

			// at last perform validation:
			validator.validate(new StreamSource(fichierxml));

		}catch (Exception ex) {
			// we are here if the document is not valid:
			// ... process validation error...
			throw new FormatValidatorException(ex.getMessage(), ex);
		}
	}

	public boolean isWellFormed(){
		try {
			validerFichierXML(file.getAbsolutePath(), false);
		}
		catch (Exception e){
			logger.warn("isWellFormed ctrl error [" + file + "]", e);
			return false;
		}
		return true;
	}

	public void checkIsWellFormed() throws NotWellFormedException{
		try{
			validerFichierXML(file.getAbsolutePath(), false);
		}
		catch(Exception e){
			throw new NotWellFormedException(this.getClass(), e.getLocalizedMessage());
		}
	}

	public String getXMLVersion(){
		String version = null;
		try  {
			DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory.newInstance();
			DocumentBuilder docBuilder = docBuilderFactory.newDocumentBuilder();
			Document doc = docBuilder.parse(this.file);
			version = doc.getXmlVersion();
		}
		catch (Exception e) {
			//do nothing
		}
		if(version==null) return  Format.NON_APPLICABLE;
		return version;
	}


	public void checkIsValid() throws NotValidException{
		try       {
			validerFichierXML(file.getAbsolutePath(), true);
		}
		catch (Exception e) {
			throw new NotValidException(this.getClass(), e.getLocalizedMessage());
		}
	}
}
