package fr.cines.pac.formatValidator.tools;

import java.io.IOException;

import org.apache.log4j.Logger;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.helpers.XMLReaderFactory;

/**
 * Gestion des fichier XML
 *
 */
public class XmlSchemaTool{

    private static Logger logger = Logger.getLogger(XmlSchemaTool.class);

    public static final String DEFAULT_PARSER_NAME = "org.apache.xerces.parsers.SAXParser";

    public XmlSchemaTool() {
    }
    /**
     * Valide un fichier XML passé en InputSource
     * @param source
     * @param validateSchemaOrDtd
     * @throws IOException
     * @throws SAXException
     */
    public void validateXML(InputSource source, boolean validateSchemaOrDtd) throws IOException, SAXException {

       // if (logger.isDebugEnabled())
            logger.debug("property:http.proxyHost=" + System.getProperty("http.proxyHost") );

        //on est dans le code de l'agent d'application donc on est sur que le jar possedant l'implementation est dans le classpath
        XMLReader parser = XMLReaderFactory.createXMLReader(DEFAULT_PARSER_NAME);

        
        if (validateSchemaOrDtd) {
            parser.setFeature("http://xml.org/sax/features/validation", true);
            parser.setFeature("http://apache.org/xml/features/validation/schema", true);
            parser.setFeature("http://apache.org/xml/features/validation/schema-full-checking", false);
        }
        else {
            parser.setFeature("http://xml.org/sax/features/validation", false);
            parser.setFeature("http://apache.org/xml/features/validation/schema", false);
            parser.setFeature("http://apache.org/xml/features/validation/schema-full-checking", false);
        }

        Validator handler = new Validator();
        parser.setErrorHandler(handler);
        parser.parse(source);
        if (handler.validationError == true) {
            throw new SAXException(handler.saxParseException.getMessage());
        }
    }


    private class Validator extends DefaultHandler
    {
        public boolean validationError = false;

  //      public String level = "";
        
        public SAXParseException saxParseException = null;

        public void error(SAXParseException exception) throws SAXException
        {
            validationError = true;
            saxParseException = exception;
        //    level = "DEBUG";// level = "ERROR";
        }

        public void fatalError(SAXParseException exception) throws SAXException
        {
            validationError = true;
            saxParseException = exception;
            //     level = "DEBUG";//  level = "FATAL";
        }

        public void warning(SAXParseException exception) throws SAXException
        {
            validationError = true;
            saxParseException = exception;
            //   level = "DEBUG";//  level = "WARN";
        }
    }
}